import sys
import pyfsdb.tools.pdbsplitter

def main():
    sys.stderr.write("dbsplitter is obsolete; please use pdbsplitter instead\n")
    pyfsdb.tools.pdbsplitter.main()


if __name__ == '__main__':
    main()



#!/usr/bin/python3

"""db2tex converts any FSDB file into a latex table.
WARNING: very little escaping is done -- watch out for mallicious input files."

import argparse
import sys
import pyfsdb

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=__doc__,
                                     epilog="Exmaple Usage: db2tex -c col1 col2 -p cc input.fsdb")

    parser.add_argument("-p", "--tabular-profile", type=str,
                        help="The column profile to pass to tabular.  The default will be all 'l's.")

    parser.add_argument("-c", "--columns", type=str, nargs="*",
                        help="Column names to include; will use all if not specified")

    parser.add_argument("-C", "--caption", type=str,
                        help="Use this as the caption for the table")

    parser.add_argument("-l", "--label", type=str,
                        help="Add a label to the table (eg: tab:foo)")

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input FSDB file")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="The output file to print latex table data to")

    args = parser.parse_args()
    return args

def latex_escape(value):
    return str(value).replace("\\","\\\\").replace("_", "\\_").replace("&","\\&")

def main():
    args = parse_args()

    inh = pyfsdb.Fsdb(file_handle = args.input_file)
    outh = args.output_file

    columns = args.columns
    if not columns:
        columns = inh.column_names

    if args.tabular_profile:
        specifier = args.tabular_profile
    else:
        specifier = "l" * len(columns)

    column_numbers = inh.get_column_numbers(columns)

        

    # write out the header info
    outh.write("\\begin{table}\n")
    outh.write("  \\begin{tabular}{%s}\n" % (specifier))

    for num, column in enumerate(columns):
        if num == 0:
            outh.write("    \\textbf{%s}" % (latex_escape(column)))
        else:
            outh.write(" & \\textbf{%s}" % (latex_escape(column)))
    outh.write(" \\\\\n")

    for row in inh:
        for num, column in enumerate(column_numbers):
            if num == 0:
                outh.write("    %s" % (latex_escape(row[column])))
            else:
                outh.write(" & %s" % (latex_escape(row[column])))
        outh.write(" \\\\\n")


    outh.write("  \\end{tabular}\n")
    if args.caption:
        outh.write("  \\caption{%s}\n" % (args.caption))
    if args.label:
        outh.write("  \\label{%s}\n" % (args.label))
    outh.write("\\end{table}\n")

if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""Converts a JSON file containing either an array of dictionaries or
individual  dictionary lines into an FSDB file"""

import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import json
import pyfsdb

def parse_args():
    """Parse command line arguments"""
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__)

    parser.add_argument("input_file", type=FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input file (json file) to read")

    parser.add_argument("output_file", type=FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="The output file (FSDB file) to write back out")

    args = parser.parse_args()
    return args

def handle_rows(out_fsdb, rows, columns):
    "Output each row in an array to the output fsdb file"
    for row in rows:
        out = []
        for column in columns:
            if column in row:
                out.append(row[column])
            else:
                out.append('')
        out_fsdb.append(out)

def json_to_fsdb(input_file, output_file):
    """A function that converts an input file stream of json dictionary
    to an output FSDB file, where the header column names are pulled
    from the first record keys."""
    first_line = next(input_file)

    try:
        rows = json.loads(first_line)
        if not isinstance(rows, list):
            rows = [rows]
    except Exception as exp:
        sys.stderr.write("failed to parse the first line as json:\n")
        sys.stderr.write(first_line)
        sys.stderr.write(str(exp))
        sys.exit(1)

    columns = sorted(list(rows[0].keys()))
    out_fsdb = pyfsdb.Fsdb(out_file_handle=output_file)
    out_fsdb.out_column_names = columns
    handle_rows(out_fsdb, rows, columns)

    for line in input_file:
        try:
            rows = json.loads(line)
            if not isinstance(rows, list):
                rows = [rows]
            handle_rows(out_fsdb, rows, columns)
        except Exception as exp:
            sys.stderr.write("failed to parse: " + line)

def main():
    "CLI wrapper around json_to_fsdb"
    args = parse_args()
    json_to_fsdb(args.input_file, args.output_file)

if __name__ == "__main__":
    main()



import sys
import pyfsdb.tools.pdbzerofill

def main():
    sys.stderr.write("dbzerofill is obsolete; please use pdbzerofill instead\n")
    pyfsdb.tools.pdbzerofill.main()


if __name__ == '__main__':
    main()



#!/usr/bin/python3

"""Splits a single FSDB file into multiple files."""

import argparse
import sys
import pyfsdb

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    parser.add_argument("-c", "--columns", type=str, nargs="*",
                        help="Columns to use when splitting the file.  If not specified, all columns are split individually.  If specified, multiple columns per file can be specified using comma separated COL1,COL2 column groupings.")

    parser.add_argument("-k", "--columns-to-keep", type=str, nargs="*",
                        help="Columns to keep in every file (timestamps, keys, etc)")

    parser.add_argument("-o", "--output-file-format", default="fsdb-split-%s.fsdb", type=str,
                        help="The output file format to use.  It should contain a percent s in it, which will be replaced by the column name being split (with commas from -c being replaced by _s).")

    parser.add_argument("-b", "--write-blanks", action="store_true",
                        help="Write out blank values as well; normally data is only written when primary columns are non-blank.")



    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input FSDB file to split")


    args = parser.parse_args()
    return args

def main():
    args = parse_args()

    input_f = pyfsdb.Fsdb(file_handle = args.input_file)

    if args.columns:
        column_names = args.columns
    else:
        column_names = input_f.column_names

    if args.columns_to_keep:
        keep_column_numbers = input_f.get_column_numbers(args.columns_to_keep)
    else:
        args.columns_to_keep = []
        keep_column_numbers = []

    split_handles = {}
    for col in column_names:
        if col in args.columns_to_keep:
            continue

        sub_columns = col.split(",")
        file_column_list = "_".join(sub_columns)

        out_column_names = []
        out_column_names.extend(args.columns_to_keep)
        out_column_names.extend(sub_columns)

        out_column_numbers = []
        out_column_numbers.extend(keep_column_numbers)
        out_column_numbers.extend(input_f.get_column_numbers(sub_columns))

        split_handles[file_column_list] = {
            "f": pyfsdb.Fsdb(out_file = (args.output_file_format % file_column_list)),
            "out_column_numbers": out_column_numbers,
            "out_column_names": out_column_names,
            "out_column_check_colnums": input_f.get_column_numbers(sub_columns),
        }
        split_handles[file_column_list]['f'].out_column_names = out_column_names

    for row in input_f.next_as_array():
        for outcol in split_handles:

            newrow = []

            # make sure we're not writing blank rows
            if not args.write_blanks:
                appends = False
                for col in split_handles[outcol]['out_column_check_colnums']:
                    if row[col] != "":
                        appends = True

                if not appends:
                    continue

            # add the existing logs
            for colnum in split_handles[outcol]['out_column_numbers']:
                newrow.append(row[colnum])
                    
            split_handles[outcol]['f'].append(newrow)


if __name__ == "__main__":
    main()



import sys
import pyfsdb.tools.pdbfullpivot

def main():
    sys.stderr.write("dbfullpivot is obsolete; please use pdbfullpivot instead\n")
    pyfsdb.tools.pdbfullpivot.main()


if __name__ == '__main__':
    main()



#!/usr/bin/python3

"""dbsum will take two matricies in FSDB and add the data together
based on particular keys.  This is similar to joining and then merging
the columns, but in one application without the need to specify column
renaming/etc with otherwise identically formatted files.

Note: this requires memory large enough to store the final results in
memory at once.
""" 

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import sys
import pyfsdb


def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
                            epilog="Exmaple Usage: dbsum -k colone -o output.fsdb input1.fsdb input2.fsdb ...")

    parser.add_argument("-k", "--keys", default=["key"], nargs="*",
                        type=str,
                        help="Keys to use for selecting columns to add")

    parser.add_argument("-c", "--columns", default=["value"], nargs="*",
                        help="Value columns to add together")

    parser.add_argument("-s", "--subtract", action="store_true",
                        help="Subtract the other files from the first")

    parser.add_argument("-o", "--output-file", type=FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="Where to save the data to")

    parser.add_argument("input_files", type=FileType('r'),
                        nargs='+', default=sys.stdin,
                        help="Input files to read from")

    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    data = {}
    keys = args.keys
    values = args.columns
    subtract = args.subtract

    first_file = True

    for input_file in args.input_files:
        with pyfsdb.Fsdb(file_handle=input_file,
                         return_type=pyfsdb.RETURN_AS_DICTIONARY) as fh:
            for row in fh:
                # loop through the keys and find the right deep tree spot
                ptr = data
                for key in keys:
                    if row[key] not in ptr:
                        ptr[row[key]] = {}
                    ptr = ptr[row[key]]

                # for each column, add it or subtract it from the results
                for value in values:
                    if value not in ptr:
                        ptr[value] = 0

                    if subtract and not first_file:
                        ptr[value] -= float(row[value])
                    else:
                        ptr[value] += float(row[value])
        first_file = False

    # save the output
    oh = pyfsdb.Fsdb(out_file_handle=args.output_file)
    oh.out_column_names = keys + values

    def write_recursive(spot, depth, indexes):
        if depth == 0:
            oh.append(indexes + list(spot.values()))
        else:
            for keyvalue in spot:
                write_recursive(spot[keyvalue], depth-1, indexes + [keyvalue])

    write_recursive(data, len(keys), [])


if __name__ == "__main__":
    main()




#!/usr/bin/python3

"""dbnormalize takes an input file and takes each column value from a
number of columns and divides it by the maximum value seen in all the
columns.

Note: this is the maximum value of all columns provided; if you want
per-column normalization, run the tool multiple times instead.

Note: this requires reading the entire file into memory.
"""

import pyfsdb
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import sys


def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
                            epilog="Exmaple Usage: dbnormalize -k column -- infile outfile")

    parser.add_argument("-k", "--keys", default=["key"], nargs="+", type=str,
                        help="The columns/keys to normalize across")

    parser.add_argument("input_file", type=FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input file to read")

    parser.add_argument("output_file", type=FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="Where to write the results")

    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    fh = pyfsdb.Fsdb(file_handle=args.input_file,
                     out_file_handle=args.output_file)
    df = fh.get_pandas()
    maxval = df[args.keys].max().max()
    for key in args.keys:
        df[key] = df[key] / maxval
    fh.put_pandas(df)
    fh.close()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""This script takes all the data in a file, and passes it to a
   jinja2 template with each row being stored in a `rows` variable.

   Note: all the rows must be loaded into memory at once.
"""

import argparse
import sys
import os

import pyfsdb
import jinja2

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__, epilog="Example: pdbj2 -j report.j2 input.fsdb output.txt")

    parser.add_argument("-j", "--jinja2-template",
                        type=argparse.FileType('r'),
                        help="The jinja2 template file to use")

    parser.add_argument("-i", "--include-file-path", type=str,
                        help="Path to allow including files from")

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs="?", default=sys.stdin,
                        help="The input file to use")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="Where to write the results to")

    args = parser.parse_args()

    if not args.jinja2_template:
        sys.stderr.write("A jinja2 template argument (-j) is required\n")
        exit(1)

    return args


def process(input_file_handle, jinja2_template, output_file_handle,
            include_file_path=None):
    "Process an input data file file and template into an output file"
    # load the data
    inh = pyfsdb.Fsdb(file_handle=input_file_handle,
                      return_type=pyfsdb.RETURN_AS_DICTIONARY)
    rows = inh.get_all()

    # get jinja2 setup
    jinja_template_data = jinja2_template.read()
    loader = None

    # allowing including of other files?
    if include_file_path:
        if include_file_path[-1] != "/":
            include_file_path += "/"   # think required?
        loader = jinja2.FileStreamLoader(include_file_path)

    # create the actual template
    template = jinja2.Environment(loader=loader)
    template = template.from_string(jinja_template_data)

    # call j2 and write the results out to the file
    output_file_handle.write(template.render({'rows': rows}))


def main():
    args = parse_args()
    process(args.input_file, args.jinja2_template, args.output_file)


if __name__ == "__main__":
    main()




#!/usr/bin/python3

"""Converts a bro (zeek) log to a file readable by FSDB.
   Bro logs are already tab separated, so we really just replace
   the headers and re-print the rest.  brotofsdb assumes
   the bro log is properly formatted (ie, tab separated already)."""

import argparse
import sys

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    parser.add_argument("-l", "--leave-bro-headers", action="store_true",
                        help="Leave the bro headers in place right after the new FSDB header")

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input file (bro log) to read")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="The output file (FSDB log) to write back out")

    args = parser.parse_args()
    return args

def main():
    args = parse_args()

    leave_headers = []

    # read in all the headers, looking for certain things
    column_names = []
    for line in args.input_file:
        if line[0] != "#":
            break


        if args.leave_bro_headers:
            leave_headers.append(line)

        if line[0:7] == "#fields":
            column_names = line.replace(".", "_").split("\t")
            column_names.pop(0)

    # print out the FSDB header
    args.output_file.write("#fsdb -F t " + " ".join(column_names))

    # optionally add back in the bro headers
    if args.leave_bro_headers:
        args.output_file.write("".join(leave_headers))

    # copy out the rest of thefile
    args.output_file.write(line)
    for line in args.input_file:
        args.output_file.write(line)

    # append our trailing command
    args.output_file.write("# " + sys.argv[0] + "\n")
    
        
if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""Passes the requested columns (-k) through the python regex escaping function.

Note: because -k can take multiple columns, input files likely need to appear
after the "--" argument-stop-parsing string.
"""

import pyfsdb
from re import escape

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import sys

def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
	                        epilog="Exmaple Usage: dbrequote -k column1 column2 -- file.fsdb")

    parser.add_argument("-k", "--keys-to-escape", type=str, nargs="+",
                        help="The keys to regexp quote")

    parser.add_argument("input_file", type=FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input file to parse")

    parser.add_argument("output_file", type=FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="Where to send the output")

    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    fs = pyfsdb.Fsdb(file_handle = args.input_file,
                     out_file_handle = args.output_file)

    convert_cols = fs.get_column_numbers(args.keys_to_escape)

    for row in fs:
        for column in convert_cols:
            row[column] = escape(row[column])
        fs.append(row)

if __name__ == "__main__":
    main()



import sys
import pyfsdb.tools.pdbheatmap

def main():
    sys.stderr.write("dbheatmap is obsolete; please use pdbheatmap instead\n")
    pyfsdb.tools.pdbheatmap.main()


if __name__ == '__main__':
    main()



import sys
import pyfsdb.tools.pdbaugment

def main():
    sys.stderr.write("dbaugment is obsolete; please use pdbaugment instead\n")
    pyfsdb.tools.pdbaugment.main()


if __name__ == '__main__':
    main()



#!/usr/bin/python3

"""dbdatetoepoch converts a timestamp column with a human date to a
unix epoch timestamp column"""

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import sys
import pyfsdb
from dateparser import parse

def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
	                        epilog="Exmaple Usage: dbdatetoepoch -d human_column -t timestamp_column input.fsdb output.fsdb")

    parser.add_argument("-d", "--date-column", default="date", type=str,
                        help="Date column to use")

    parser.add_argument("-t", "--timestamp-column", default="timestamp", type=str,
                        help="Column to create for storing an epoch column")

    parser.add_argument("input_file", type=FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="")

    parser.add_argument("output_file", type=FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="")

    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    fh = pyfsdb.Fsdb(file_handle=args.input_file,
                     out_file_handle=args.output_file)
    column_names = fh.column_names
    fh.out_column_names = column_names + [args.timestamp_column]

    date_column = fh.get_column_number(args.date_column)
    for row in fh:
        timestamp_value = 0
        try:
            timestamp_value = parse(row[date_column]).timestamp()
        except Exception:
            pass
        row[-1] = timestamp_value  # XXX: this should be append
        fh.append(row)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""Fills a column with zeros if it was otherwise blank"""

import sys
import argparse
import pyfsdb

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    parser.add_argument("-v", "--value", default="0", type=str,
                        help="Fill columns with this value")

    parser.add_argument("-c", "--columns", type=str, nargs="+",
                        help="Fill these columns")

    parser.add_argument("-k", "--key-column", default="timestamp", type=str,
                        help="Use this column as the timestamp/key column to increment")

    parser.add_argument("-b", "--bin-size", default=1, type=int,
                        help="Bin-size to check for missing rows")

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="")

    args = parser.parse_args()

    if args.columns == None:
        sys.stderr.write("The --columns argument is required\n")
        exit(1)

    return args

def main():
    args = parse_args()

    fh = pyfsdb.Fsdb(file_handle=args.input_file,
                     out_file_handle=args.output_file)

    store_columns = fh.get_column_numbers(args.columns)
    time_column = fh.get_column_number(args.key_column)
    value = args.value
    bin_size = args.bin_size

    last_index = None

    for row in fh:
        if last_index == None:
            # first row, just store it
            last_index = int(row[time_column])
        elif last_index != int(row[time_column]):
            for skipped_time in range(last_index + bin_size, int(row[time_column]), bin_size):
                newrow = list(row)
                newrow[time_column] = str(skipped_time)
                for column in store_columns:
                    newrow[column] = value
                fh.append(newrow)
            last_index = int(row[time_column])
        fh.append(row)

    fh.write_finish()

if __name__ == "__main__":
    main()



#!/bin/python3

"""dbreversepivot takes an input file with time/value columns, and 
   pivots the table into a narrow table with one line per old column.

   For example, if the input was this:

       #fsdb -F s time foo bar
       1 10 0
       2 30 20
       3 0 40

   It would convert this to:

       #fsdb -F s time key value
       1 foo 10
       2 bar 20
       2 foo 30
       3 bar 40

   This is the inverse operation of dbfullpivot.
   """

import sys
import argparse
import pyfsdb

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    parser.add_argument("-k", "--key-column", default="key", type=str,
                        help="The key column to use in the output for column names to store in")

    parser.add_argument("-c", "--columns", nargs="+", type=str,
                        help="The columns to pivot into keys")

    parser.add_argument("-v", "--value-column", default="value", type=str,
                        help="What output column to store the value for what was found in the columns")

    parser.add_argument("-o", "--other-columns", default=[], type=str, nargs="*",
                        help="Other columns to copy to every row")

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input FSDB file to read")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="The output FSDB file to write to")


    args = parser.parse_args()
    return args

def main():
    args = parse_args()

    # set up storage structures
    storage = {}
    columns = {}

    # from the input, get extract column numbers/names
    key_column = args.key_column
    value_column = args.value_column
    other_columns = args.other_columns
    columns = args.columns

    # open the input file stream
    input = pyfsdb.Fsdb(file_handle = args.input_file,
                        return_type=pyfsdb.RETURN_AS_DICTIONARY)
    output = pyfsdb.Fsdb(out_file_handle = args.output_file)
    output.out_column_names = [key_column, value_column] + other_columns

    # for each row, remember each value based on time and key
    for row in input:
        for column in columns:
            out_row = [column, row[column]]
            for other in other_columns:
                out_row.append(row[other])
            output.append(out_row)

    output.close()
    
if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""Converts an FSDB file to a stream of json dictionaries."""

import sys
import argparse
import json
import pyfsdb

def parse_args():
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    parser.add_argument("-a", "--array", action="store_true",
                        help="Format the output as a large array.")

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input file (FSDB file) to read")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="The output file (json file) to write back out")

    args = parser.parse_args()
    return args


def fsdb_to_json(input_file, output_file, as_array=False):
    "Converts an FSDB file to a stream of json dictionary records"
    in_fsdb = pyfsdb.Fsdb(file_handle=input_file,
                          return_type=pyfsdb.RETURN_AS_DICTIONARY)

    end_token = ""
    close_token = ""
    if as_array:
        output_file.write("[")
        end_token = ","
        close_token = "]"

    previous_record = None
    for record in in_fsdb:
        if previous_record:
            output_file.write(json.dumps(previous_record) + end_token + "\n")
        previous_record = record

    if previous_record:
        output_file.write(json.dumps(previous_record) + close_token + "\n")

def main():
    "CLI wrapper around fsdb_to_json"
    args = parse_args()
    fsdb_to_json(args.input_file, args.output_file,
                 as_array = args.array)


if __name__ == "__main__":
    main()



import sys
import pyfsdb.tools.pdbensure

def main():
    sys.stderr.write("dbensure is obsolete; please use pdbensure instead\n")
    pyfsdb.tools.pdbensure.main()


if __name__ == '__main__':
    main()



#!/usr/bin/python3

import sys
import os
import argparse
import collections

import pyfsdb


def parse_args():
    formatter_class = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=formatter_class,
                                     description=__doc__)

    parser.add_argument("-k", "--keys", default=["key"], type=str, nargs="*",
                        help="Key to use when counting for uniqueness")

    parser.add_argument("-c", "--count", action="store_true",
                        help="Count the columns")

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="")

    args = parser.parse_args()
    return args


def output_results(outh, keys, data, count=False):
    for subkey in data:
        if isinstance(data[subkey], dict):
            output_results(outh, keys + [subkey], data[subkey], count)
        else:
            if count:
                outh.append(keys + [subkey, data[subkey]])
            else:
                outh.append(keys + [subkey])


def filter_unique_columns(in_file_handle, out_file_handle, keys,
                          count=False):
    fh = pyfsdb.Fsdb(file_handle=in_file_handle)
    ofh = pyfsdb.Fsdb(out_file_handle=out_file_handle)

    key_columns = fh.get_column_numbers(keys)

    # set the output column names
    if count:
        ofh.column_names = keys +  ['count']
    else:
        ofh.column_names = keys

    num_keys = len(keys)
    if len(key_columns) == 1:
        counters = collections.Counter()
        for row in fh:
            counters[row[key_columns[0]]] += 1
    else:
        counters = {}
        for row in fh:
            pointer = counters
            for keynum in range(num_keys - 2):
                if row[key_columns[keynum]] not in pointer:
                    pointer[row[key_columns[keynum]]] = {}
                pointer = pointer[row[key_columns[keynum]]]

            if row[key_columns[num_keys - 2]] not in pointer:
                pointer[row[key_columns[num_keys - 2]]] = collections.Counter()
            pointer[row[key_columns[num_keys - 2]]][row[key_columns[num_keys - 1]]] += 1


    # output the results, with optional counts
    # (if statement at outer tier for speed)
    output_results(ofh, [], counters, count)

    ofh.close()


def main():
    args = parse_args()

    filter_unique_columns(args.input_file, args.output_file,
                          keys=args.keys, count=args.count)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""dbensure can be used that some or all fields in a table contain data.

If rows with the specified columns (default: all) don't contain data,
they're dropped from the output rows."""

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import sys

import pyfsdb

def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
	                        epilog="Exmaple Usage: dbensure input_file.fsdb output_file.fsdb")

    parser.add_argument("-c", "--columns", nargs="*",
                        help="The columns to check in the data")

    parser.add_argument("-v", "--fill", default=None, type=str,
                        help="Don't drop the rows but fill with this value if a column is missing")

    parser.add_argument("-e", "--print-error", action="store_true",
                        help="Print an error message on each dropped row")

    parser.add_argument("input_file", type=FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input file to process")

    parser.add_argument("output_file", type=FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="Where to send the output data")

    args = parser.parse_args()
    return args

def filter_row(row, columns, fill_value, print_error):
    for column in columns:
        if row[column] == "":
            if fill_value:
                row[column] = fill_value
            else:
                if print_error:
                    sys.stderr.write("# dbensure dropping row:" + str(row) + "\n")
                return
    return row

def main():
    args = parse_args()
    fh = pyfsdb.Fsdb(file_handle=args.input_file,
                     out_file_handle=args.output_file)

    if args.columns:
        column_nums = fh.get_column_numbers(args.columns)
    else:
        column_nums = list(range(len(fh.column_names)))

    fh.filter(filter_row, args=[column_nums, args.fill, args.print_error])

if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""Find the top (bottom) N rows identified by a key and sorted by a
value column.  The result will be N rows with unique keys, sorted
by the top N value columns.

Examples:

Given a dataset like:

  #fsdb -F t one two three
  10	a	42
  20	b	99
  20	a	50

and running it as:

  topn -n 2 -k two -v three

Will produce:

  #fsdb -F t one two three
  20	b	99
  20	a	50

(ie, one row for each a and b)

Or:

  topn -n 1 -k two -v three

creates:

  #fsdb -F t one two three
  20	b	99

"""

import sys
import argparse
import pyfsdb

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    parser.add_argument("-k", "--key", default="key", type=str,
                        help="Name of the key column to use")

    parser.add_argument("-v", "--value", default="value", type=str,
                        help="Name of the value column to use when sorting")

    parser.add_argument("-n", "--max-rows", default=20, type=int,
                        help="Number of rows to return")

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="")

    args = parser.parse_args()
    return args

def maybe_add_data(data_by_key, row, key_column, value_column,
                   max_value, max_rows):
    if row[key_column] in data_by_key:
        # update the existing
        data_by_key[row[key_column]] = row
    else:
        # it's a new row, append it
        data_by_key[row[key_column]] = row
        # and maybe drop an old row
        if len(data_by_key) > max_rows:
            # need to drop the lowest, which also has max_value
            delete_this = min(data_by_key,
                              key=lambda x: float(data_by_key[x][value_column]))
            del data_by_key[delete_this]

    # calculate he new minimum
    min_key = min(data_by_key,
                  key=lambda x: float(data_by_key[x][value_column]))
    max_value = float(data_by_key[min_key][value_column])

    return max_value

def main():
    args = parse_args()

    fin = pyfsdb.Fsdb(file_handle=args.input_file)
    fout = pyfsdb.Fsdb(out_file_handle=args.output_file)
    fout.column_names = fin.column_names

    (key_column, value_column) = fin.get_column_numbers([args.key, args.value])
    max_value = None
    data_by_key = {}
    data_values = []
    for row in fin:
        if row[value_column] is not None and row[value_column] is not "-" and row[value_column] is not "":
            if max_value is None or max_value < float(row[value_column]):
                max_value = maybe_add_data(data_by_key, row,
                                           key_column, value_column,
                                           max_value, args.max_rows)
            elif row[key_column] in data_by_key and float(data_by_key[row[key_column]][value_column]) < float(row[value_column]):
                max_value = maybe_add_data(data_by_key, row,
                                           key_column, value_column,
                                           max_value, args.max_rows)
            
    for key in data_by_key:
        fout.append(data_by_key[key])

if __name__ == "__main__":
    main()

    



#!/usr/bin/python3

import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
import pyfsdb


def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
                            epilog="Exmaple Usage: ")

    parser.add_argument("-c", "--columns", type=str, nargs=2,
                        help="The two columns to use for the x and y axes")

    parser.add_argument("-v", "--value-column", default="count", type=str,
                        help="The value column to plot as the heat map")

    parser.add_argument("-i", "--invert", action="store_true",
                        help="Invert the foreground/background colors")

    parser.add_argument("-F", "--add-fractions", action="store_true",
                        help="Add text fraction labels to the grid")

    parser.add_argument("-R", "--add-raw", action="store_true",
                        help="Add text raw-value labels to the grid")

    parser.add_argument("-L", "--add-labels", action="store_true",
                        help="Add x/y axis labels")

    parser.add_argument("--label-column", default=None, type=str,
                        help="Column to use for labeling the squares")

    parser.add_argument("-fs", "--font-size", default=None, type=int,
                        help="Set the fontsize for labels")

    parser.add_argument("-C", "--cmap", default="Blues_r", type=str,
                        help="matplotlib colormap to use (good choices: Blues_r, gray, PuBu_r, summer_r, YlGn_r)")

    parser.add_argument("--list-cmaps", action="store_true",
                        help="List the colormap values available for -C")

    parser.add_argument("--label-limit", default=30, type=int,
                        help="The maximum length of a label;" +
                        "  If longer, truncate with ...s in the middle. " +
                        "Use 0 if infinite is desired.")

    parser.add_argument("input_file", type=FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="Input fsdb file to read")

    parser.add_argument("output_file", type=str,
                        nargs='?', default="out.png",
                        help="Where to write the png file to")

    args = parser.parse_args()

    if args.list_cmaps:
        all = plt.colormaps()
        print("\n".join(all))
        exit()

    if not args.columns or len(args.columns) != 2:
        raise ValueError("exactly 2 columns must be passed to -c")

    return args


def maybe_shrink_label(label, length_limit=30):
    label = str(label)
    if len(label) <= length_limit:
        return label
    part_length = int((length_limit-3) / 2) # save room for middle dots
    right_len = -part_length
    return label[0:part_length+1] + "..." + label[right_len:]


def normalize(input_data, columns, value_column, label_column=None):
    """Loops over all of the rows of dict data extracting a tuple of:

     data: the data in array/dict format, normalized to MIN->1.0
     dataset: the dataset in a deep dictonary format
     min_value: the minimum value seen
     max_value: the maximum value seen
     xcols: the list of x column labels
     ycols: the list of y column labels
     labelset: the labels in a deep dictonary format if label_column was specified"""
    # loop over all the rows calculating the min/max values and save results
    min_value = None
    max_value = None
    ycols = {}  # stores each unique second value
    dataset = {}  # nested tree structure
    labelset = {}

    for row in input_data:
        if not max_value:
            max_value = float(row[value_column])
        else:
            max_value = max(max_value, float(row[value_column]))

        if not min_value:
            min_value = float(row[value_column])
        else:
            min_value = min(min_value, float(row[value_column]))

        label = None
        x_value = row[columns[0]]
        y_value = row[columns[1]]
        value = row[value_column]
        if label_column and label_column in row:
            label = row[label_column]
        if x_value not in dataset:
            dataset[x_value] = {y_value: float(value)}
        else:
            dataset[x_value][y_value] = float(value)

        # set the optional labels
        if label:
            if x_value not in labelset:
                labelset[x_value] = {y_value: label}
            else:
                labelset[x_value][y_value] = label

        ycols[y_value] = 1

    # merge the data into a two dimensional array
    data = []
    xcols = sorted(dataset.keys())
    ycols = sorted(ycols.keys())
    for first_column in xcols:
        newrow = []
        for second_column in ycols:
            if second_column in dataset[first_column] and max_value > 0:
                val = dataset[first_column][second_column] / max_value
                newrow.append(val)
            else:
                newrow.append(0.0)
        data.append(newrow)

    return {'data': data,
            'dataset': dataset,
            'min_value': min_value,
            'max_value': max_value,
            'xcols': xcols,
            'ycols': ycols,
            'labelset': labelset}


def create_heat_map(input_data, columns, value_column,
                    add_labels=False, add_raw=False,
                    add_fractions=False, invert=False,
                    font_size=None, max_label_size=20,
                    cmap='Blues_r', label_column=None):

    results = normalize(input_data, columns, value_column, label_column)

    (data, dataset, xcols, ycols, labelset) = \
        (results['data'],
         results['dataset'],
         results['xcols'],
         results['ycols'],
         results['labelset'])

    grapharray = np.array(data)
    if not invert:
        grapharray = 1 - grapharray

    # generate the graph
    fig, ax = plt.subplots()

    # set the size
    fig.set_dpi(150)
    fig.set_size_inches(16, 9)

    ax.imshow(grapharray, vmin=0.0, vmax=1.0, cmap=cmap)
    # ax.grid(ls=':')

    ax.set_xlabel(maybe_shrink_label(columns[1], max_label_size), fontsize=font_size)
    ax.set_ylabel(maybe_shrink_label(columns[0], max_label_size), fontsize=font_size)

    # note: xlabels are applied on the y tick labels
    if add_labels:
        ax.set_yticks(np.arange(len(dataset)))
        ax.set_xticks(np.arange(len(ycols)))

        xlabels = [maybe_shrink_label(x, max_label_size) for x in xcols]
        ax.set_yticklabels(xlabels, fontsize=font_size)

        ylabels = [maybe_shrink_label(y, max_label_size) for y in ycols]
        ax.set_xticklabels(ylabels, fontsize=font_size)

        plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
                 rotation_mode="anchor")

    if add_fractions:
        for i in range(len(grapharray)):
            for j in range(len(grapharray[i])):
                ax.text(j, i, "{:1.1f}".format(grapharray[i][j]),
                        ha="center", va="center", color="r",
                        fontsize=font_size)
    elif add_raw:
        for i, first_column in enumerate(xcols):
            for j, second_column in enumerate(ycols):
                try:
                    value = dataset[first_column][second_column]
                    if value != "0" and value != 0:
                        ax.text(j, i, "{:1.1f}".format(float(value)),
                                ha="center", va="center", color="r",
                                fontsize=font_size)
                except Exception:
                    pass

    elif label_column:
        for i, first_column in enumerate(xcols):
            for j, second_column in enumerate(ycols):
                try:
                    label = labelset[first_column][second_column]
                    if label:
                        ax.text(j, i, str(label),
                                ha="center", va="center", color="r",
                                fontsize=font_size)
                except Exception:
                    pass

    fig.tight_layout()
    return (fig, data, dataset)


def main():
    args = parse_args()

    # read in the input data
    f = pyfsdb.Fsdb(file_handle=args.input_file,
                    return_type=pyfsdb.RETURN_AS_DICTIONARY)

    (fig, data, dataset) = \
        create_heat_map(f, args.columns, args.value_column,
                        args.add_labels, args.add_raw,
                        args.add_fractions, args.invert,
                        args.font_size, args.label_limit,
                        args.cmap, args.label_column)

    fig.savefig(args.output_file,
                bbox_inches="tight", pad_inches=0)

    # import pprint
    # pprint.pprint(dataset)


if __name__ == "__main__":
    main()



"""Creates a CDF fraction column from another column's data.  Note
that this requires loading all the data into memory for efficiency."""

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
from logging import debug, info, warning, error, critical
import logging
import sys
import pyfsdb

def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
                            epilog="Exmaple Usage: ")

    parser.add_argument("-c", "--data-column", default="data",
                        type=str,
                        help="The input column name to create a CDF from")

    parser.add_argument("-C", "--cdf-column", default=None,
                        type=str,
                        help="The output CDF column name -- if none, '_cdf' will be appended to --column")

    parser.add_argument("-R", "--raw-column", default=None, type=str,
                        help="Output raw accumulating count to this column")

    parser.add_argument("-P", "--percent-column", default=None, type=str,
                        help="Output a percentage column, in addition to the CDF column")

    parser.add_argument("-F", "--fraction-column", default=None, type=str,
                        help="Output a percentage column, in addition to the CDF column")

    parser.add_argument("--log-level", default="info",
                        help="Define the logging verbosity level.")

    parser.add_argument("input_file", type=FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input FSDB file")

    parser.add_argument("output_file", type=FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="Where to write the output FSDB to")

    args = parser.parse_args()
    log_level = args.log_level.upper()
    logging.basicConfig(level=log_level,
                        format="%(levelname)-10s:\t%(message)s")
    return args

def process_cdf(input_file, output_file, data_column,
                out_cdf=None, raw_column=None,
                fraction_column=None, percent_column=None,
                use_max=False):
    # open input and output fsdb handles
    fh = pyfsdb.Fsdb(file_handle=input_file)
    oh = pyfsdb.Fsdb(out_file_handle=output_file)

    # crate output columns
    out_columns = fh.column_names

    # Add the CDF column
    if not out_cdf:
        out_cdf = data_column + "_cdf"
    
    out_columns.append(out_cdf)

    # add the raw and percentage columns
    if raw_column:
        out_columns.append(raw_column)
    if fraction_column:
        out_columns.append(fraction_column)
    if percent_column:
        out_columns.append(percent_column)

    oh.out_column_names = out_columns

    df = fh.get_pandas(data_has_comment_chars=True)

    denominator = df[data_column].sum()
    df[out_cdf] = df[data_column].cumsum() / denominator
    if raw_column:
        df[raw_column] = df[data_column].cumsum()
    if fraction_column:
        df[fraction_column] = df[data_column] / denominator
    if percent_column:
        df[percent_column] = 100.0 * df[data_column] / denominator

    oh.put_pandas(df)

def main():
    args = parse_args()

    process_cdf(args.input_file, args.output_file,
                args.data_column, args.cdf_column,
                args.raw_column, args.fraction_column,
                args.percent_column)

if __name__ == "__main__":
    main()



import sys
import pyfsdb.tools.pdbdatetoepoch

def main():
    sys.stderr.write("dbdatetoepoch is obsolete; please use pdbdatetoepoch instead\n")
    pyfsdb.tools.pdbdatetoepoch.main()


if __name__ == '__main__':
    main()



import sys
import pyfsdb.tools.pdbkeyedsort

def main():
    sys.stderr.write("dbkeyedsort is obsolete; please use pdbkeyedsort instead\n")
    pyfsdb.tools.pdbkeyedsort.main()


if __name__ == '__main__':
    main()



#!/usr/bin/python3

"""dbdatetoepoch converts a unix epoch timestamp column into a human
readable date string usting strftime with an adjustable format."""

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import sys
import pyfsdb
from dateparser import parse
import time

def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
	                        epilog="Exmaple Usage: dbdatetoepoch -d human_column -t timestamp_column input.fsdb output.fsdb")

    parser.add_argument("-t", "--timestamp-column", default="timestamp",
                        type=str, help="Column to use with the epoch timestamp")

    parser.add_argument("-T", "--time-column", default="timestamp_human", type=str,
                        help="The output time/date column to create")

    parser.add_argument("-f", "--format", default="%Y-%m-%d %H:%M", type=str,
                        help="The output format to use in the time column")

    parser.add_argument("input_file", type=FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="")

    parser.add_argument("output_file", type=FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="")

    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    fh = pyfsdb.Fsdb(file_handle=args.input_file,
                     out_file_handle=args.output_file,
                     converters={args.timestamp_column: float})
    column_names = fh.column_names
    fh.out_column_names = column_names + [args.time_column]

    timestamp_column = fh.get_column_number(args.timestamp_column)

    colfmt = args.format

    for row in fh:
        row[-1] = time.strftime(colfmt, time.localtime(row[timestamp_column]))
        fh.append(row)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""This script expects to augment a FSDB stream (or file) with
   additional information found in another FSDB file.  The second file
   is loaded entirely into memory in order to accomplish this.  Any
   row in the stream file that has exactly matching keys in the second
   file will be augmented to using data from the 'values' columns is
   the augment file.

   This duplicates dbjoin to a large extent, but dbaugment is faster
   when one side is small because it avoids sorting.  dbaugment can
   also operate on streaming data, since sorting isn't required.
"""

import argparse
import sys
import os

import pyfsdb

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    parser.add_argument("-k", "--keys", default=['timestamp', 'key'], type=str, nargs="+",
                        help="Keys to match against")

    parser.add_argument("-v", "--values", type=str, nargs="*",
                        help="Value columns to insert on a match")

    parser.add_argument("-V", "--include-all-values", action="store_true",
                        help="Include all the value columns from the matching row")

    parser.add_argument("-i", "--include-unmatched-augment-rows", action="store_true",
                        help="Include (at the bottom) any rows from the augment file that failed to match any stream rows.")


    parser.add_argument("stream_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="")

    parser.add_argument("augment_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="")

    args = parser.parse_args()

    if (not args. values or len(args.values) == 0) \
       and not args.include_all_values:
            sys.stderr.write("Either -v or -V is required\n")
            exit(1)

    return args

def dump_remaining(fsh, struct, empty_num, key_cols, value_cols):
    """Adds remaining unseen records in the stored augment file into the
    output handle.  Recursion used to descend a deep-encoded
    structure.
    """
    if 'data' in struct:
        # end condition
        if 'used' not in struct:
            # unused row found,
            # construct the output row, consisting of keys
            # followed by blanks, followed by values
            # XXX: this breaks when the keys aren't the first columns
            row = []
            for key in key_cols:
                row.append(struct['data'][key])
            row.extend([''] * empty_num)
            for value in value_cols:
                row.append(struct['data'][value])

            # append the created row
            fsh.append(row)
        return
    else:
        # we're not fully deep in the tree, keep diving
        for item in struct:
            dump_remaining(fsh, struct[item], empty_num,
                           key_cols, value_cols)

def main():
    args = parse_args()

    # read in the augument file entirely first
    augh = pyfsdb.Fsdb(file_handle = args.augment_file,
                       return_type=pyfsdb.RETURN_AS_DICTIONARY)

    # store each row based on its list of keys, but storing
    # each additional key as a deeper layer of dictionaries.
    # the final key used is 'data' to store the data itself
    savestruct = {}
    for row in augh:
        current = savestruct
        # traverse/create the nested structure
        for key in args.keys:
            if row[key] not in current:
                current[row[key]] = {}
            current = current[row[key]]
        current['data'] = row

    # read in stream file, and augment each row with the new columns
    streamh = pyfsdb.Fsdb(file_handle = args.stream_file)

    # determine which columns need to be added on (potentially all)
    if not args.values or len(args.values) == 0:
        columns = augh.column_names
        args.values = []
        for column in columns:
            if column not in args.keys:
                args.values.append(column)

    # create the output stream to store the data
    outh = pyfsdb.Fsdb(out_file_handle = args.output_file)
    outh.out_column_names = streamh.column_names + args.values

    key_columns = streamh.get_column_numbers(args.keys)

    for row in streamh:
        current = savestruct
        # traverse/create the nested structure
        for key in key_columns:
            if row[key] not in current:
                current = None
                break
            current = current[row[key]]
    
        # on a deep match finding
        if current:
            current['used'] = 1
            current = current['data']
            for value in args.values:
                row.append(current[value])

        outh.append(row)

    # Now loop through all data adding any rows 
    if args.include_unmatched_augment_rows:
        dump_remaining(outh, savestruct,
                       # total original column numbers - key count
                       len(streamh.column_names) - len(args.keys),
                       args.keys, args.values)

if __name__ == "__main__":
    main()



"""Plots a ROC curve from a given FSDB file with a confidence and true column

This requires both matplotlib and sklearn to function.
"""

import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve, plot_roc_curve, auc
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
from logging import debug, info, warning, error, critical
import logging
import sys
import pyfsdb

def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
                            epilog="Exmaple Usage: pdbroc in_file.fsdb out_file.png")

    parser.add_argument("-t", "--truth-column", default="true", type=str,
                        help="Specifies the column name that marks the true rows")

    parser.add_argument("-T", "--truth-value", default="1", type=str,
                        help="Specifies the truth column value that should be used to indicate 'true'")

    parser.add_argument("-c", "--confidence-column", default="confidence", type=str,
                        help="The column name holding the confidence value")

    parser.add_argument("-i", "--invert-false-truths", action="store_true",
                        help="If specified, a 0 confidence for a false row will be considered a true negative (and a good classification) -- otherwise the confidence is supposed to respect the confidence of the truth value, higher always being better.")

    parser.add_argument("-l", "--label", default="ROC", type=str,
                        help="The label to put on the legend")

    parser.add_argument("-a", "--output-auc", action="store_true",
                        help="Output the AUC value to the terminal as well")

    parser.add_argument("--log-level", default="info",
                        help="Define the logging verbosity level (debug, info, warning, error, fotal, critical).")

    parser.add_argument("input_file", type=FileType('r'), nargs='?', default=sys.stdin,
                        help="The input FSDB file to plot")

    parser.add_argument("output_file", type=str, nargs='?', default="",
                        help="The output (PNG) file to write; if not specified plot in a window; if NONE don't plot at all (useful to just print AUC)")

    args = parser.parse_args()
    log_level = args.log_level.upper()
    logging.basicConfig(level=log_level,
                        format="%(levelname)-10s:\t%(message)s")
    return args


# taken from
# https://www.codespeedy.com/how-to-plot-roc-curve-using-sklearn-library-in-python/
def plot_roc(fpr, tpr, label="ROC"):
    plt.plot([0, 1], [0, 1], linestyle=':', color='black')
    plt.plot(fpr, tpr, color='blue', label=label)
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.legend()
    return plt


def main():
    args = parse_args()

    inh = pyfsdb.Fsdb(file_handle=args.input_file,
                      return_type=pyfsdb.RETURN_AS_DICTIONARY)

    confidence_column = args.confidence_column
    true_column = args.truth_column
    true_value = args.truth_value

    trues = []
    confidence = []
    for row in inh:
        if row[true_column] == true_value:
            trues.append(1)
            confidence.append(float(row[confidence_column]))
        else:
            trues.append(0)
            if args.invert_false_truths:
                confidence.append(1.0-float(row[confidence_column]))
            else:
                confidence.append(float(row[confidence_column]))

    (tpr, fpr, thresholds) = roc_curve(trues, confidence)
    plot_roc(fpr, tpr, label=args.label)

    if args.output_auc:
        print(f"AUC: {auc(fpr, tpr)}")

    if args.output_file:
        if args.output_file != "NONE":
            plt.savefig(args.output_file)
    else:
        plt.show()


if __name__ == "__main__":
    main()




#!/usr/bin/python3

"""Sort "mostly sorted" large FSDB files using a double pass

dbkeyedsort reads a file twice, sorting the data by the column specified via the -c/--column option.  During the first pass, it counts all the rows per key to manage which lines it needs to memorize as it's making its second pass.  During the second pass, it only stores in memory the lines that are out of order.  This can greatly optimize the amount of memory stored when the data is already in a fairly sorted state (which is common for the output of map/reduce operations such as hadoop).  This comes at the expense of needing to read the entire dataset twice, which means its impossible to use `stdin` to pass in data; instead a filename must be specified instead.  The output, though, may be `stdout`.

USAGE

    dbkeyedsort <-c COLUMN> INPUT_FILE OUTPUT_FILE

EXAMPLE

    dbkeyedsort -c timestamp sortthis.fsdb output.fsdb

"""

import sys
import argparse
import pyfsdb

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    parser.add_argument("-c", "--column", type=str, 
                        help="The column to sort by")

    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Output number of rows cached to stderr")

    parser.add_argument("input_file", type=str,
                        help="The file to read (can't be stdin; must be seekable)")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="The file to write to")

    args = parser.parse_args()

    if 'column' not in args:
        raise ValueError('--column or -c is required')

    return args

def main():
    args = parse_args()

    f = pyfsdb.Fsdb(args.input_file)
    key_column = f.get_column_number(args.column)
    key_counts = {}

    line_count = 0
    cached_count = 0

    # memorize all the keys and the number of rows in each
    for row in f.next_as_array():
        line_count += 1
        if row[key_column] not in key_counts:
            key_counts[row[key_column]] = 1
        else:
            key_counts[row[key_column]] += 1

    key_list = list(key_counts.keys())
    key_list.sort()

    # re-open the input file to re-read
    f = pyfsdb.Fsdb(args.input_file, out_file_handle=args.output_file)

    stored_lines = {}

    # memorize all the keys and the number of rows in each
    current_key = key_list.pop(0)
    for row in f.next_as_array():
        if row[key_column] != current_key:
            # the current lines are arriving too early; cache them
            cached_count += 1
            if row[key_column] not in stored_lines:
                stored_lines[row[key_column]] = [row]
            else:
                stored_lines[row[key_column]].append(row)
        else:
            f.append(row)
            key_counts[current_key] -= 1

            while key_counts[current_key] == 0:
                #import pdb; pdb.set_trace()
                # we're done with this key list
                if len(key_list) == 0:
                    break # done!

                # grab a new key
                current_key = key_list.pop(0)

                # write out any cached lines
                if current_key in stored_lines:
                    for stored_row in stored_lines[current_key]:
                        f.append(stored_row)
                        key_counts[current_key] -= 1
                    del stored_lines[current_key]

    f.write_finish()

    if args.verbose:
        sys.stderr.write("cached %d/%d lines\n" % (cached_count, line_count))

if __name__ == "__main__":
    main()




import sys
import pyfsdb.tools.pdbnormalize

def main():
    sys.stderr.write("dbnormalize is obsolete; please use pdbnormalize instead\n")
    pyfsdb.tools.pdbnormalize.main()


if __name__ == '__main__':
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyfsdb",
    version="1.2.5",
    author="Wes Hardaker",
    author_email="opensource@hardakers.net",
    description="A python implementation of the flat-file streaming database",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gawseed/pyfsdb",
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            # migrating to pdb prefixes
            'pdbtopn = pyfsdb.tools.pdbtopn:main',
            'pdbaugment = pyfsdb.tools.pdbaugment:main',
            'bro2fsdb = pyfsdb.tools.bro2fsdb:main',
            'pdbcoluniq = pyfsdb.tools.pdbcoluniq:main',
            'pdbfullpivot = pyfsdb.tools.pdbfullpivot:main',
            'pdbreversepivot = pyfsdb.tools.pdbreversepivot:main',
            'pdbzerofill = pyfsdb.tools.pdbzerofill:main',
            'pdbkeyedsort = pyfsdb.tools.pdbkeyedsort:main',
            'pdbsplitter = pyfsdb.tools.pdbsplitter:main',
            'json2fsdb = pyfsdb.tools.json2fsdb:main',
            'fsdb2json = pyfsdb.tools.fsdb2json:main',
            'fsdb2many = pyfsdb.tools.fsdb2many:main',
            'pdb2tex = pyfsdb.tools.pdb2tex:main',
            'pdbformat = pyfsdb.tools.pdbformat:main',
            'pdbreescape = pyfsdb.tools.pdbreescape:main',
            'pdbensure = pyfsdb.tools.pdbensure:main',
            'pdbheatmap = pyfsdb.tools.pdbheatmap:main',
            'pdbdatetoepoch = pyfsdb.tools.pdbdatetoepoch:main',
            'pdbepochtodate = pyfsdb.tools.pdbepochtodate:main',
            'pdbnormalize = pyfsdb.tools.pdbnormalize:main',
            'pdbsum = pyfsdb.tools.pdbsum:main',
            'pdbj2 = pyfsdb.tools.pdbj2:main',
            'pdbcdf = pyfsdb.tools.pdbcdf:main',
            'pdbroc = pyfsdb.tools.pdbroc:main',

            # obsolete wrappers
            'dbtopn = pyfsdb.obsolete.dbtopn:main',
            'dbaugment = pyfsdb.obsolete.dbaugment:main',
            'dbcoluniq = pyfsdb.obsolete.dbcoluniq:main',
            'dbfullpivot = pyfsdb.obsolete.dbfullpivot:main',
            'dbreversepivot = pyfsdb.obsolete.dbreversepivot:main',
            'dbzerofill = pyfsdb.obsolete.dbzerofill:main',
            'dbkeyedsort = pyfsdb.obsolete.dbkeyedsort:main',
            'dbsplitter = pyfsdb.obsolete.dbsplitter:main',
            'db2tex = pyfsdb.obsolete.db2tex:main',
            'dbformat = pyfsdb.obsolete.dbformat:main',
            'dbreescape = pyfsdb.obsolete.dbreescape:main',
            'dbensure = pyfsdb.obsolete.dbensure:main',
            'dbheatmap = pyfsdb.obsolete.dbheatmap:main',
            'dbdatetoepoch = pyfsdb.obsolete.dbdatetoepoch:main',
            'dbsum = pyfsdb.obsolete.dbsum:main',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires = '>=3.6',
    test_suite='nose.collector',
    tests_require=['nose'],
)



import sys
import pyfsdb.tools.pdbreversepivot

def main():
    sys.stderr.write("dbreversepivot is obsolete; please use pdbreversepivot instead\n")
    pyfsdb.tools.pdbreversepivot.main()


if __name__ == '__main__':
    main()



#/usr/bin/python

"""
pyfsdb.Fsdb - class for reading and writing Fsdb files

# creating a Fsdb object attached to a file to read:
f = pyfsdb.Fsdb("data.fsdb")

# Getting column numbers for named columns:
name_col = f.get_column_number('name')
time_col = f.get_column_number('time')

# iterating over a fsdb file:
for data in f:
    print(data[name_col] + ".." + data[time_col])

# iterating over a fsdb file with dictionaries:
# (slower)
f = pyfsdb.Fsdb("data.fsdb", return_type=pyfsdb.RETURN_AS_DICTIONARY)
for data in f:
    print(data['name'] + ".." + data['time'])

# writing FSDB formatted data out
# (note: a single Fsdb object can be set up for both reading and writing)
f = pyfsdb.Fsdb(out_file = "my.fsdb")
f.out_column_names = ['foo','bar','baz']
f.out_separator = "\t"
f.append(["a", "b", "c"])
f.close() # closes and writes trailing comments

# (if the FSDB object is also reading a file, default values
# for out_column_names and out_separator will be taken from the input file).

# Convenience wrappers

# foreach: return a list of the values from the second column
f = pyfsdb.Fsdb("data.fsdb")
results = f.foreach(lambda x: x[2])

# filter: reads the input file, filters it using myfilt, and writes to output
def myfilt(row):
    return [int(row['mycol']) * 2]    

f = pyfsdb.Fsdb("data.fsdb", out_file="output.fsdb")
results = f.filter(myfilt)

"""

import sys
if sys.version_info[0] < 3:
    raise Exception("Must be using Python 3")

RETURN_AS_ARRAY = 1
RETURN_AS_DICTIONARY = 2

class Fsdb(object):
    """Reads FSDB files from the perl FSDB module.

       (see the fsdb module documentation for full details)
       """

    fileh = None
    _header_line = None
    current_line = None
    filename = None
    _separator = "\t"
    _separator_token = "t"
    _headers = None
    _column_names = {}
    _pass_comments = True
    _mapping = None
    _have_read_header = False

    _out_file = None
    _out_file_handle = None
    _out_header_line = None
    _out_column_names = []
    _out_column_names_set = False
    _out_separator = "\t"
    _out_separator_token = "t"
    _out_command_line = "____BROKEN____" # ick, magic
    _save_command_history = True

    def __init__(self,
                 filename = None,
                 file_handle = None,
                 return_type=RETURN_AS_ARRAY,
                 out_file = None,
                 out_file_handle = None,
                 pass_comments = 'y',
                 out_command_line = "____INTERNAL____",
                 write_nones_as_blanks = True,
                 column_names=None,
                 converters=None,
                 save_command_history=True):
        """Returns a Fsdb class that can be used as an iterator.

           return_type can be pyfsdb.RETURN_AS_ARRAY (default) or
           RETURN_AS_DICTIONARY to return dictionary based rows with 
           indexes as columns (this is slower).

           If `pass_comments` is True and both an input and output file
           handle are available, any comments read in while reading
           are printed to the output.

           `converters` may be passed in as an array or dict of
           converters to call (such as int, float, etc)
        """

        self.return_type = return_type
        self.filename = filename
        self.fileh = file_handle
        self._column_names = column_names
        self._pass_comments = pass_comments
        self._write_nones_as_blanks = write_nones_as_blanks
        self._header_written = False
        self._converters = converters
        self._save_command_history = save_command_history

        if pass_comments not in ['y', 'n', 'e']:
            raise ValueError("pass_comments must be y/n/e")

        if out_file:
            self.out_file = out_file
        elif out_file_handle:
            self._out_file_handle = out_file_handle
        self._maybe_open_out()

        if out_command_line == "____INTERNAL____":
            self.out_command_line = " ".join(sys.argv)
        else:
            self.out_command_line = out_command_line

        self._comments = []

    @property
    def file_handle(self):
        "The input file handle being read."
        return self.fileh

    @file_handle.setter
    def file_handle(self, fileh):
        self.fileh = fileh

        return fileh

    @property
    def headers(self):
        "Headers for the file handle being read."
        self.maybe_open_filehandle()
        if not self._headers:
            self.read_header()
        return self._headers

    @headers.setter
    def headers(self, value):
        self._headers = value

    @property
    def header_line(self):
        "The top #fsdb header line in the file being read."
        self.maybe_open_filehandle()
        if not self._header_line:
            self.read_header()
        return self._header_line

    @header_line.setter
    def header_line(self, value):
        self._header_line = value

    @property
    def column_names(self):
        "An array of column names for the file being read"
        self.maybe_open_filehandle()
        if not self._column_names:
            self.read_header()
        return list(self._column_names.keys())

    @column_names.setter
    def column_names(self, values):
        mapping = self.__create_column_name_mapping__(values)
        self._header_line = \
            self.create_header_line(columns = values,
                                    separator_token = self._separator_token)
        self.headers = self._header_line

    @property
    def separator(self):
        """The 'separator_token' is the argument that comes after -F in the 
           fsdb header and the separator is it's translation;
           eg, for tab-based separator the separator_token would 
           be the 't' character and the separator would be '\t'.

           Changing this will also change the stored separator_token value."""
        self.maybe_open_filehandle()
        if not self._separator:
            self.read_header()
        return self._separator

    @separator.setter
    def separator(self, value):
        self._separator = value
        self._separator_token = self.convert_separator_token(value)
        self._header_line = self.create_header_line(columns = self.column_names, separator_token = self._separator_token)

    @property
    def separator_token(self):
        """The separator_token character for the file being read or written.

           The 'separator_token' is the argument that comes after -F in the 
           fsdb header; eg, for a tab-based separator the separator_token would 
           be the 't' character and the separator would be '\t'.

           Changing this will also change the stored separator value."""
        self.maybe_open_filehandle()
        if not self._separator_token:
            self.read_header()
        return self._separator_token

    @separator_token.setter
    def separator_token(self, value):
        self._separator_token = value
        self._separator = self.parse_separator(self.separator_token)
        self._header_line = self.create_header_line(columns = self.column_names, separator_token = self._separator_token)

    def __create_column_name_mapping__(self, columns):
        """Internal

        Creates a list of columns and numbers for rapid mapping
        at the start to make more rapid lookups later.
        """
        mapping = {'names': {}, 'numbers': {},
                   'header': { 'separator': self.separator}}

        argn = 0
        for token in columns:
            mapping['names'][token] = argn
            mapping['numbers'][argn] = token
            argn += 1

        self._column_names = mapping['names']
        self.column_nums = mapping['numbers']

        if not self._out_column_names_set:
            self._out_column_names = self._column_names.keys()

        self._mapping = mapping
        return mapping

    #
    # output properties
    #
    @property
    def out_file(self):
        """The output file being written to (if one is being written)"""
        return self._out_file

    @out_file.setter
    def out_file(self, value):
        self._out_file = value
        self._out_file_handle = None
        self._maybe_open_out()

    @property
    def out_file_handle(self):
        """The output file being written to (if one is being written)"""
        return self._out_file_handle

    @out_file_handle.setter
    def out_file_handle(self, value):
        self._out_file_handle = value
        self._out_file = None
        self._maybe_open_out()

    def _maybe_open_out(self):
        if not self._out_file_handle and self.out_file:
            self._out_file_handle = open(self.out_file, "w")
        elif self.out_file_handle:
            try:
                self._out_file = self._out_file_handle.name
            except:
                pass

    @property
    def out_separator(self):
        """The separator for the output.

           Changing this will also change the stored
           out_separator_taken value.

           This should not be changed after the header has 
           already been written.
        """
        return self._out_separator

    @out_separator.setter
    def out_separator(self, value):
        self._out_separator = value
        self._out_separator_token = self.convert_separator_token(value)
        self._out_header_line = self.create_header_line()

    @property
    def out_separator_token(self):
        """The separator for the output.

           Changing this will also change the stored
           out_separator value.

           This should not be changed after the header has 
           already been written.
        """
        return self._out_separator_token

    @out_separator_token.setter
    def out_separator_token(self, value):
        self._out_separator_token = value
        self._out_separator = self.parse_separator(self._out_separator_token)
        self._out_header_line = self.create_header_line()

    @property
    def out_header_line(self):
        "The top #fsdb header line to write to the output."
        self.maybe_open_filehandle()
        self._out_header_line = self.create_header_line()
        return self._out_header_line

    @out_header_line.setter
    def out_header_line(self, value):
        """The output header line to print when writing.

        This must be set prior to writing the first row for it to
        be written"""
        self._out_header_line = value

    @property
    def out_command_line(self):
        """The output trailing command to print as the last line.

         The out_command_line is printed with a '#   | ' prefix 
         to preserve the history of the command run.  It defaults
         to " ".join(sys.argv).  Set to None if you wish to surpress
         printing of the line entirely."""
        return self._out_command_line

    @out_command_line.setter
    def out_command_line(self, value):
        self._out_command_line = value

    @property
    def out_column_names(self):
        """An array of column names for the output file being written.

        This must be set prior to writing the first row, and modifies
        the internal output_header_line too."""
        if len(self._out_column_names) == 0:
            self._out_column_names = self.column_names
        return self._out_column_names

    @out_column_names.setter
    def out_column_names(self, values):
        mapping = self.__create_column_name_mapping__(values)
        self.out_column_names_set = True
        self._out_header_line = self.create_header_line(values)

    # support functions

    def create_header_line(self, columns = None, separator_token = None):
        "Returns a header string for the stored column_names and separator/separator_token."
        if not columns:
            columns = self.out_column_names

        if separator_token is None:
            separator_token = self.out_separator_token

        if separator_token is None:
            separator_token = self.separator_token

        # create the header line
        header_line = "#fsdb -F " + separator_token + " " + " ".join(columns) + "\n"
        self._have_read_header = True

        return header_line

    @property
    def converters(self):
        """The list of conversion routines.

        It may be an array, with a converter per column,
        or a dict with a convert per named column.

        This must be set before the file is opened/read.

        Useful converted may include int, float, etc.
        """
        return self._converters

    @converters.setter
    def converters(self, new_converters):
        self._converters = new_converters

    def _convert_converters(self):
        if isinstance(self._converters, dict):
            converters = []
            # convert this to an array based on column names
            for column_name in self.column_names:
                if column_name in self._converters and \
                   callable(self._converters[column_name]):
                    converters.append(self._converters[column_name])
                else:
                    converters.append(None)
            self._converters = converters

    # column accessor helpers
    def get_column_number(self, column_name):
        "Given a column_name, returns its integer index into an array of values."
        self.maybe_open_filehandle()
        self.read_header()
        return self._column_names[column_name]

    def get_column_numbers(self, column_names):
        "Given a list of column_names, returns a list of integer index into an array of values."
        self.maybe_open_filehandle()
        self.read_header()
        column_numbers = []
        for name in column_names:
            column_numbers.append(self.get_column_number(name))
        return column_numbers

    def get_column_name(self, column_number):
        "Given an integer column number, returns its column name."
        self.maybe_open_filehandle()
        self.read_header()
        return self.column_nums[column_number]

    def set_iterator_function(self):
        "XXX: change this to an property"
        if self.return_type == RETURN_AS_DICTIONARY:
            self.__next__ = self._next_as_dict
        else:
            self.__next__ = self._next_as_array

    def __iter__(self):
        """Returns an iterator object for looping over an fsdb file."""
        if not self.filename and not self.file_handle:
            raise ValueError("No filename or handle currently available for reading")
        # XXX: throw error on -1 parse
        return self

    def maybe_open_filehandle(self, mode="r"):
        "Internal"
        if not self.file_handle and self.filename:
            self.file_handle = open(self.filename, mode)

        return self.file_handle

    def __next__(self):
        """Returns the next array of data from an fsdb file.
           Returns an array by default, or a dictionary if return_type 
           was set to pyfsdb.RETURN_AS_DICTIONARY."""

        fh = self.maybe_open_filehandle()
        if not self._header_line:
            self.read_header()

        self._convert_converters()

        if self.return_type == RETURN_AS_DICTIONARY:
            self.__next__ = self._next_as_dict
        else:
            self.__next__ = self._next_as_array

        if not fh:
            return None

        return self.__next__()

    def _handle_comment(self, line):
        """Handle a comment by printing it, possibly with header init first,
        and then returning the next line in the file"""
        if self._pass_comments != 'n' and self._out_file_handle:

            if self.append != self._append_really:
                # we haven't printed anything yet, so we haven't written
                # the fsdb_header yet

                self._write_header_line()

            if self._pass_comments == 'y':
                self._out_file_handle.write(line)
            else:
                self._comments.append(line)
        return next(self.fileh)

    def _convert_array_values(self, row):
        for (n, converter) in enumerate(self._converters):
            if converter is not None and row[n] is not None:
                row[n] = converter(row[n])
        return row

    def _next_as_array(self):
        """Return the next object as an array of columns."""

        line = next(self.fileh)
        while line and line[0] == '#':
            line = self._handle_comment(line)

        # return an array of data
        self.current_line = line
        self._current_row = line.rstrip("\n\r").split(self.separator)
        if len(self._current_row) < len(self._column_names):
            n = (len(self._column_names))-len(self._current_row)
            self._current_row.extend([''] * n)
        if self._converters:
            self._current_row = self._convert_array_values(self._current_row)
        return self._current_row

    def _next_as_dict(self):
        """Return the next object as a dictionary, with column 
        names as the indexes.  

        Note: This is less efficient than returning data as an
        array using the normal __next__() routine."""

        array = self._next_as_array()

        return_dict = {}
        for index in range(0, len(array)):
            return_dict[self.column_nums[index]] = array[index]

        return return_dict

    # generator type function for returning a row as an array
    def next_as_array(self):
        """Generator function to return a row as an array.

        Using a generator is faster than using the Fsdb object
        as a iterator."""

        fh = self.maybe_open_filehandle()

        try:
            line = next(self.fileh)
            while line:
                while line and line[0] == '#':
                    line = self._handle_comment(line)

                # return an array of data
                self.current_line = line
                self._current_row = line.rstrip("\n\r").split(self.separator)
                yield self._current_row
                line = next(self.fileh)
        except StopIteration as e:
            return

    def next_as_dict(self):
        """Generator function to return an array row.

        Using a generator is faster than using the Fsdb object
        as a iterator."""

        fh = self.maybe_open_filehandle()

        try:
            line = next(self.fileh)
            while line:
                while line and line[0] == '#':
                    line = self._handle_comment(line)

                # return an array of data
                self.current_line = line
                self._current_row = line.rstrip("\n\r").split(self.separator)

                return_dict = {}
                for index in range(0, len(self._current_row)):
                    return_dict[self.column_nums[index]] = self._current_row[index]

                yield return_dict
                line = next(self.fileh)
        except StopIteration as e:
            return

    def parse_separator(self, separator = None):
        """Converts a separator ("t") into a separator_token ("\t") """
        if separator == "t":
            return "\t"
        elif separator == "S":
            return "  "
        elif separator == "s":
            return " "
        elif separator[0] == "c":
            return separator[1:]
        elif separator[0] == "C":
            return separator[1:] # won't handle multiples like manual says
        elif separator[0] == "x":
            return chr(int("0x" + separator[1:],0))
        elif separator[0] == "X": # won't handle multiples like manual says
            return chr(int("0x" + separator[1:],0))
        elif separator == "D":
            # python NONE to splits on all white space
            return None

        # unknown
        raise ValueError("Unknown separator value: " + separator)

    def convert_separator_token(self, separator_token = None):
        """Converts a separator ("\t") into a separator_token ("t") """
        if separator_token == "\t":
            return "t"
        elif separator_token == "  ":
            return "S"
        elif separator_token == " ":
            return "s"
        elif len(separator_token) == 1:
            return "C" + separator_token
        elif separator_token == None:
            # XXX
            separator_token = 'D'

        # unknown
        raise ValueError("Unknown separator token value: " + separator_token)

    def read_header(self, line = None):
        """Internal

        Returns a dict of header -> column numbers.

        The header line should be in the form:

            #fsdb -option value column1(separator)column2...

        Returns:
           [0, {
                  names: { colname: colnum, ...},
                  numbers: { colnum: colname, ...}
                  header: { separator: separator_string}
               }]    on success
           [-1, "error description"]         on failure
        """

        if self._have_read_header:
            return [0, self._mapping]
        self._have_read_header = True

        if not line:
            line = next(self.file_handle)
        self._header_line = line
        self._headers = [self._header_line]

        args = line.split()
        if args[0] != "#fsdb":
            raise ValueError("failed to find expected #fsdb header")

        # should we use argparse here?
        argn = 1
        separator = None  # (D)efault is all white space
        while args[argn][0] == '-':
            if args[argn] == "-F":
                argn += 1
                self._separator_token = args[argn]
            else:
                return [-1, "Unown option: " + args[argn]]

            argn += 1

        self._separator = self.parse_separator(self._separator_token)

        # join the remainder of the arguments back together to split
        # by the correct separator

        # XXX: allow other separators in the header
        remainder = " ".join(args[argn:]).rstrip()
        args = remainder.split(" ")

        mapping = self.__create_column_name_mapping__(args)
        self._separator = mapping['header']['separator']

        return [0, mapping]

    def row_as_string(self, row = None):
        """Converts an array row to an FSDB output line."""
        if not row:
            row = self._current_row
        return self.separator.join(row) + "\n"

    #
    # Useful higher-level functions
    #

    def get_all(self):
        """Read all the data into memory and return it as an array of rows."""
        all_rows = []
        for row in self:
            all_rows.append(row)
        return all_rows

    def get_pandas(self, usecols=None, comment="#",
                   data_has_comment_chars=False, **kwargs):
        """Returns a pandas dataframe for the given data.  Warning: this
        cannot preserve comments in the files; FSDB comments are
        stripped from the output.  Any other args will be passed to
        pandas.read_csv()
        """
        import pandas
        column_names = self.column_names # forces opening and reading headers

        if data_has_comment_chars:
            # when data has the comment character in the middle,
            # pandas.read_csv fails because it assumes comment
            # characters are line breaks.  We have no choice but to
            # read all the data ourselves (slow).
            slow_data = self.get_all()

            df = pandas.DataFrame(slow_data, columns=column_names)

            for column in column_names:
                try:
                    df[column] = pandas.to_numeric(df[column])
                except:
                    pass

            return df
        else:
            return pandas.read_csv(self.file_handle, sep='\t', comment=comment,
                                   names=column_names, usecols=usecols,
                                   **kwargs)

    def put_pandas(self, df):
        "saves a pandas dataframe to the output file"
        import pandas
        if not self._out_column_names:
            self.out_column_names = df.columns.values.tolist()
        self._write_header_line()
        df.to_csv(self._out_file_handle, sep="\t", header=False,
                  index=False)

    def foreach(self, fn, return_results=True, args=[]):
        """Applies a function fn to each row, returning an 
        aggregate list of results if desired."""
        results = []
        if return_results:
            for row in self:
                results.append(fn(row, *args))
        else:
            # do this separately for speed purposes
            for row in self:
                fn(row, *args)

        return results

    def filter(self, fn, args=[]):
        """Applies a function to the input stream rows, and saves the output
        results back to the output fsdb handle."""
        for row in self:
            filtered = fn(row, *args)
            if filtered:
                self.append(filtered)

    #
    # writing functions
    #

    def append(self, row = None):
        """Writes a passed in row (or the one previously read) to the output file."""
        self._append_init(row)

    def extend(self, rows = None):
        """Writes multiple rows to the output FSDB"""
        for row in rows:
            self.append(row)

    def _write_header_line(self):
        if self._out_header_line:
            self._out_file_handle.write(self._out_header_line)
        elif self._header_line:
            # assuming copy the original
            self._out_file_handle.write(self._header_line)

        # if we write the header line,
        #    it's now ok to set the output function to the right function
        self.append = self._append_really
        self._header_written = True

    def _append_init(self, row = None):
        # internallly, if we haven't written the header out we do that first
        # then switch our operator for speed

        self._write_header_line()

        self.append(row)

    def _append_really(self, row = None):
        if not row:
            row = self._current_row
        if isinstance(row, dict):
            row = [row[x] for x in self.out_column_names]
        if self._write_nones_as_blanks:
            for i in range(0,len(row)):
                if row[i] == None:
                    row[i] = ''
        self._out_file_handle.write(self._out_separator.join(map(str,row)) + "\n")
    # backwards compatible ... don't use
    def write_row(self, row = None):
        self.append(row)

    def write_finish(self):
        self.close()

    def close(self):
        """Writes final processing command comment to the output file and closes it."""
        if self.fileh:
            self.fileh.close()
            self.fileh = None

        if self._out_file_handle:
            # ignore closing errors
            try:
                if not self._header_written:
                    self._write_header_line()
                if self._save_command_history and self.out_command_line:
                    for comment_line in self._comments:
                        self._out_file_handle.write(comment_line)
                    self._out_file_handle.write("#   | " + self.out_command_line + "\n")
                self._out_file_handle.close()
            except:
                pass
            self._out_file_handle = None

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()





import sys
import pyfsdb.tools.pdbformat

def main():
    sys.stderr.write("dbformat is obsolete; please use pdbformat instead\n")
    pyfsdb.tools.pdbformat.main()


if __name__ == '__main__':
    main()



import sys
import pyfsdb.tools.pdbreescape

def main():
    sys.stderr.write("dbreescape is obsolete; please use pdbreescape instead\n")
    pyfsdb.tools.pdbreescape.main()


if __name__ == '__main__':
    main()



import sys
import pyfsdb.tools.pdbcoluniq

def main():
    sys.stderr.write("dbcoluniq is obsolete; please use pdbcoluniq instead\n")
    pyfsdb.tools.pdbcoluniq.main()


if __name__ == '__main__':
    main()



#!/usr/bin/python3

"""Outputs a python-string formatted line for every input FSDB row,
with column names acting as variables into the format string."""

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import sys
import pyfsdb

def parse_args():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description=__doc__,
	                        epilog="Example: dbformat -f 'I can print {col1} and {col2}'")

    parser.add_argument("-f", "--format", type=str,
                        help="The python-based format string to use")

    parser.add_argument("input_file", type=FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="The input FSDB file to read")

    parser.add_argument("output_file", type=FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="The output text file to write to")

    args = parser.parse_args()

    if not args.format:
        sys.stderr.write("-f is a required argument\n")
        exit(1)
        
    return args

def main():
    args = parse_args()

    inh = pyfsdb.Fsdb(file_handle = args.input_file,
                      return_type=pyfsdb.RETURN_AS_DICTIONARY)
    outh= args.output_file

    format_string = args.format

    for row in inh:
        outh.write(format_string.format(**row) + "\n")

if __name__ == "__main__":
    main()



#!/bin/python3

"""dbfullpivot takes an input file with time/key/value pairs, and 
   pivots the table into a wide table with one new column per key value.

   For example, if the input was this:

       #fsdb -F s time key value
       1 foo 10
       2 bar 20
       2 foo 30
       3 bar 40

   It would convert this to:

       #fsdb -F s time foo bar
       1 10 0
       2 30 20
       3 0 40

   """

import sys
import argparse
import pyfsdb

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="")

    parser.add_argument("output_file", type=argparse.FileType('w'),
                        nargs='?', default=sys.stdout,
                        help="")

    parser.add_argument("-k", "--key-column", default="key", type=str,
                        help="The key to pivot around")

    parser.add_argument("-t", "--time-column", default="time", type=str,
                        help="The binned time column to use")

    args = parser.parse_args()
    return args

def main():
    args = parse_args()

    # set up storage structures
    storage = {}
    columns = {}

    # open the input file stream
    input = pyfsdb.Fsdb(file_handle = args.input_file)

    # from the input, get extract column numbers/names
    time_column = input.get_column_number(args.time_column)
    key_column = input.get_column_number(args.key_column)
    column_names = input.column_names

    # for each row, remember each value based on time and key
    for row in input:
        # if the time hasn't been seen before, allocate the sub-structure
        if row[time_column] not in storage:
            storage[row[time_column]] = {}

        for column_num in range(0,len(row)):
            # remember all values of non-time and non-key columns 
            if column_num != time_column and column_num != key_column:
                storage[row[time_column]][row[key_column]] = row[column_num]
                # record that we've seen this column before
                columns[row[key_column]] = 1

    # open the output stream, and set it's properties
    out = pyfsdb.Fsdb(out_file_handle = args.output_file)

    # the output columns will be a merge of the time column, and
    # previously seen key-index values.
    output_columns = [args.time_column]
    output_columns.extend(columns.keys())
    out.out_column_names = output_columns

    # Output all data, grouped by time_key 
    for time_key in storage:

        # create a row containing a column for every seen key
        row = [time_key]
        for column in columns:
            if column not in storage[time_key] or storage[time_key][column] == "":
                row.append("0")
            else:
                row.append(storage[time_key][column])

        # write it out
        out.append(row)

    
if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""fsdb2many converts a single FSDB file into many, by creating
other file names based on a column of the original."""

import sys
import argparse
import pyfsdb
import re

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=__doc__,
                                     epilog="fsdb2many -c key -o outputdir/%s.fsdb mybigfile.fsdb")

    parser.add_argument("-c", "--column", default="key", type=str,
                        help="Column to split on")

    parser.add_argument("-o", "--output-pattern",
                        default="fsdb2many-out-%s.fsdb",
                        type=str,
                        help="Output pattern to split on, which should contain a PERCENT S to use for inserting the column value being saved to that file.")

    parser.add_argument("input_file", type=argparse.FileType('r'),
                        nargs='?', default=sys.stdin,
                        help="str")

    args = parser.parse_args()

    return args

def main():
    args = parse_args()

    # open the input file
    inh = pyfsdb.Fsdb(file_handle=args.input_file)
    key_column = inh.get_column_number(args.column)

    out_handles = {}

    for row in inh:
        value = row[key_column]

        # see if we have an open file handle for this one yet
        if value not in out_handles:
            # new value, so open a new file handle to save data for it
            file_name = re.sub("[^-.0-9a-zA-Z_]", "_", str(value))
            outh = pyfsdb.Fsdb(out_file=(args.output_pattern % file_name))
            outh.column_names = inh.column_names
            out_handles[value] = outh

        # save the row to the file based on its value
        out_handles[value].append(row)

    # clean up
    for handle in out_handles:
        out_handles[handle].close()

if __name__ == "__main__":
    main()
